package main

import (
	"gorm.io/gorm"
	"reflect"
	"testing"
	"time"
)

var sampleDB *gorm.DB

func SampleSetUpDatabase() error {
	sampleDB = GetConnection()

	// Check the connection
	_, err := sampleDB.DB()
	if err != nil {
		return err
	}

	return nil
}

func TestSampleDatabaseConnectionAndAutoMigrate(t *testing.T) {
	err := SampleSetUpDatabase()
	if err != nil {
		t.Fatalf("failed to connect to database: %v", err)
	}

	sampleDB = GetConnection()
	err = sampleDB.Migrator().AutoMigrate(&User{}, &Calendar{}, &Appointment{}, &TaskList{})
	if err != nil {
		t.Fatalf("failed to auto-migrate: %v", err)
	}
}

func TestSampleUserTable(t *testing.T) {
	expectedFields := map[string]reflect.Kind{
		"Model":     reflect.Struct,
		"Username":  reflect.String,
		"FirstName": reflect.String,
		"LastName":  reflect.String,
		"Calendar":  reflect.Struct,
	}

	user := User{}
	userType := reflect.TypeOf(user)

	for i := 0; i < userType.NumField(); i++ {
		field := userType.Field(i)
		fieldName := field.Name
		fieldType := field.Type.Kind()

		expectedType, ok := expectedFields[fieldName]
		if !ok {
			t.Errorf("Unexpected field name: %s", fieldName)
		}
		if fieldType != expectedType {
			t.Errorf("Field %s has unexpected type: %s", fieldName, fieldType)
		}
	}
}

func TestSampleCalendarTable(t *testing.T) {
	expectedFields := map[string]reflect.Kind{
		"Model":        reflect.Struct,
		"Name":         reflect.String,
		"UserID":       reflect.Uint,
		"Appointments": reflect.Slice,
	}

	calendar := Calendar{}
	userType := reflect.TypeOf(calendar)

	for i := 0; i < userType.NumField(); i++ {
		field := userType.Field(i)
		fieldName := field.Name
		fieldType := field.Type.Kind()
		expectedType, ok := expectedFields[fieldName]
		if !ok {
			t.Errorf("Unexpected field name: %s", fieldName)
		}
		if fieldType != expectedType {
			t.Errorf("Field %s has unexpected type: %s", fieldName, fieldType)
		}
	}
}

func TestSampleAppointmentTable(t *testing.T) {
	expectedFields := map[string]reflect.Kind{
		"Model":       reflect.Struct,
		"Subject":     reflect.String,
		"Description": reflect.String,
		"StartTime":   reflect.Struct,
		"Length":      reflect.Uint,
		"OwnerID":     reflect.Uint,
		"OwnerType":   reflect.String,
		"Attendees":   reflect.Slice,
	}

	appointment := Appointment{}
	userType := reflect.TypeOf(appointment)

	for i := 0; i < userType.NumField(); i++ {
		field := userType.Field(i)
		fieldName := field.Name
		fieldType := field.Type.Kind()
		expectedType, ok := expectedFields[fieldName]
		if !ok {
			t.Errorf("Unexpected field name: %s", fieldName)
		}
		if fieldType != expectedType {
			t.Errorf("Field %s has unexpected type: %s", fieldName, fieldType)
		}
	}
}

func TestSampleTaskListTable(t *testing.T) {
	expectedFields := map[string]reflect.Kind{
		"Model":        reflect.Struct,
		"Appointments": reflect.Slice,
	}

	taskList := TaskList{}
	userType := reflect.TypeOf(taskList)

	for i := 0; i < userType.NumField(); i++ {
		field := userType.Field(i)
		fieldName := field.Name
		fieldType := field.Type.Kind()
		expectedType, ok := expectedFields[fieldName]
		if !ok {
			t.Errorf("Unexpected field name: %s", fieldName)
		}
		if fieldType != expectedType {
			t.Errorf("Field %s has unexpected type: %s", fieldName, fieldType)
		}
	}

	// SampleTearDown()
}

func TestSampleUserCalendarRelationship(t *testing.T) {
	user := GetUserWithCalendar()
	if err := db.Create(&user).Error; err != nil {
		t.Fatalf("Failed to create user: %v", err)
	}

	var retrievedUser User
	if err := db.Preload("Calendar").First(&retrievedUser, user.ID).Error; err != nil {
		t.Fatalf("Failed to retrieve user: %v", err)
	}

	if retrievedUser.Calendar.ID != user.Calendar.ID {
		t.Errorf("User's calendar ID doesn't match expected value")
	}
	if retrievedUser.Calendar.Name != user.Calendar.Name {
		t.Errorf("User's calendar name doesn't match expected value")
	}
}

func TestSampleCalendarAppointmentRelationship(t *testing.T) {
	user := GetUserWithCalendar()
	if err := db.Create(&user).Error; err != nil {
		t.Fatalf("Failed to create user: %v", err)
	}

	calendar := Calendar{
		Name:   "QCalendar",
		UserID: user.ID,
	}
	if err := db.Create(&calendar).Error; err != nil {
		t.Fatalf("Failed to create calendar: %v", err)
	}
	appointment := GetAppointmentWithCalender(calendar)
	if err := db.Create(&appointment).Error; err != nil {
		t.Fatalf("Failed to create appointment: %v", err)
	}

	var retrievedAppointment Appointment
	if err := db.First(&retrievedAppointment, appointment.ID).Error; err != nil {
		t.Fatalf("Failed to retrieve appointment: %v", err)
	}
	if retrievedAppointment.OwnerID != calendar.ID || retrievedAppointment.OwnerType != "Calendar" {
		t.Errorf("Appointment doesn't belong to the correct calendar")
	}

	var retrievedCalendar Calendar
	if err := db.Preload("Appointments").First(&retrievedCalendar, user.Calendar.ID).Error; err != nil {
		t.Fatalf("Failed to retrieve calendar: %v", err)
	}

}

func TestSampleTaskListAppointmentRelationship(t *testing.T) {
	taskList := TaskList{}
	if err := db.Create(&taskList).Error; err != nil {
		t.Fatalf("Failed to create task list: %v", err)
	}

	appointment := GetAppointmentWithTaskList(taskList)
	if err := db.Create(&appointment).Error; err != nil {
		t.Fatalf("Failed to create appointment: %v", err)
	}

	var retrievedAppointment Appointment
	if err := db.First(&retrievedAppointment, appointment.ID).Error; err != nil {
		t.Fatalf("Failed to retrieve appointment: %v", err)
	}
	if retrievedAppointment.OwnerID != taskList.ID || retrievedAppointment.OwnerType != "TaskList" {
		t.Errorf("Appointment doesn't belong to the correct calendar")
	}

}

func TestSampleUserAppointmentRelationship(t *testing.T) {
	user1 := GetUser("younesious")
	user2 := GetUser("roozbehious")
	db.Create(&user1)
	db.Create(&user2)

	appointment := Appointment{
		Subject:     "Team Meeting",
		Description: "Discuss progress on project X",
		StartTime:   time.Now(),
		Length:      60,
		OwnerID:     user1.ID,
		OwnerType:   "User",
	}
	db.Create(&appointment)

	appointment.Attendees = []User{user2}
	db.Save(&appointment)

	var retrievedAppointment Appointment
	if err := db.Preload("Attendees").First(&retrievedAppointment, appointment.ID).Error; err != nil {
		t.Fatalf("Failed to retrieve appointment: %v", err)
	}
	if len(retrievedAppointment.Attendees) != 1 {
		t.Errorf("Expected 1 attendee, but got %d", len(retrievedAppointment.Attendees))
	}
	if retrievedAppointment.Attendees[0].Username != "roozbehious" {
		t.Errorf("Expected attendee username to be 'roozbehious', but got '%s'", retrievedAppointment.Attendees[0].Username)
	}

	var retrievedUser1 User
	if err := db.Preload("Calendar.Appointments").First(&retrievedUser1, user1.ID).Error; err != nil {
		t.Fatalf("Failed to retrieve user1: %v", err)
	}

	var retrievedUser2 User
	if err := db.Preload("Calendar.Appointments").First(&retrievedUser2, user2.ID).Error; err != nil {
		t.Fatalf("Failed to retrieve user2: %v", err)
	}
}

/*
func SampleTearDown() {
	sqlDB, _ := db.DB()
	defer func(sqlDB *sql.DB) {
		err := sqlDB.Close()
		if err != nil {
			panic(err)
		}
	}(sqlDB)
}
*/
